"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.GlobalObjects import GlobalObjectManager

class Constant(float):
    """
    Constant parameter that can be used by all neurons and synapses.

    The class ``Constant`` derives from ``float``, so any legal operation on floats (addition, multiplication) can be used.

    If a Neuron/Synapse defines a parameter with the same name, the constant parameters will not be visible.

    Example:

    ```python

    tau = ann.Constant('tau', 20)
    factor = ann.Constant('factor', 0.1)
    real_tau = ann.Constant('real_tau', tau*factor)

    neuron = ann.Neuron(
        equations='''
            real_tau*dr/dt + r =1.0
        '''
    )
    ```

    The value of the constant can be changed anytime with the ``set()`` method. Assignments will have no effect (e.g. ``tau = 10.0`` only creates a new float).

    The value of constants defined as combination of other constants (``real_tau``) is not updated if the value of these constants changes (changing ``tau`` with ``tau.set(10.0)`` will not modify the value of ``real_tau``).

    :param name: name of the constant (unique), which can be used in equations.
    :param value: the value of the constant, which must be a float, or a combination of Constants.
    """
    def __new__(cls, name, value):
        return float.__new__(cls, value)
        
    def __init__(self, name, value):
        """
        Constructor, implicitly calls __new__()
        """
        self.name = name
        "Name."
        self.value = value
        "Value."

        GlobalObjectManager().add_constant(self)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.__str__()

    def set(self, value:float, network=None) -> None:
        """
        Changes the value of the constant.

        :param value: Value.
        :param network: Network instance which should be updated. By default, all active networks are updated.
        """
        self.value = value

        # change for all active networks
        if network is None:
            for net_id in NetworkManager()._get_network_ids():
                if NetworkManager().is_compiled(net_id=net_id):
                    getattr(NetworkManager().cy_instance(net_id=net_id), '_set_'+self.name)(self.value)

        # update an individual network
        else:
            if NetworkManager().is_compiled(net_id=network.id):
                getattr(NetworkManager().cy_instance(net_id=network.id), '_set_'+self.name)(self.value)
            else:
                raise ValueError("Setting a constant for individual networks is only allowed afer compilation.")



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
# (directly imported by CodeGenerator if needed)
additional_global_functions = ""

launch_config = {
    'init': """
        _threads_per_block = 0;
        _nb_blocks = 0;

    #ifdef _DEBUG
        std::cout << "Kernel configuration is a fixed 2D kernel" << std::endl;
    #endif
""",
    'update': """
        // Generate the kernel launch configuration
        _threads_per_block = 0;
        _nb_blocks = 0;

    #ifdef _DEBUG
        std::cout << "Kernel configuration is a fixed 2D kernel" << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'global': {
        'parameter': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
""",
        'variable': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
"""
    }
}


attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'global': {
        'parameter': """
        // Global %(attr_type)s %(name)s
        %(name)s = 0.0;
    """,
        'variable': """
        // Global %(attr_type)s %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
"""
    }
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global %(attr_type)s %(name)s
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s - host
        %(name)s.clear();
        %(name)s.shrink_to_fit();

        // %(name)s - device
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->num_rows_ * this->num_columns_ * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->num_rows_ * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            //cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, num_rows_ * num_columns_ * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': """
            // %(name)s: semiglobal
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            //cudaMemcpy( %(name)s.data(), gpu_%(name)s, post_ranks_.size() * sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
""",
    'global': """
            // %(name)s: global
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            //cudaMemcpy( &%(name)s, gpu_%(name)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
"""
}

delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",

        'pyx_struct':
"""
        # Uniform delay
        int delay""",
        'init': """
    delay = delays[0][0];
""",
        'pyx_wrapper_init':
"""
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor':
"""
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    }
}

spike_event_transmission = {
    'device_kernel': """// gpu device kernel for projection %(id_proj)s
__global__ void cu_proj%(id_proj)s_psp( const long int t, const %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* num_events, %(conn_args_header)s %(kernel_args_header)s ) {
    int tid = threadIdx.x;
    int col_idx = blockIdx.x * blockDim.x + tid;

    __shared__ %(float_prec)s shared_psp[32];
    while( col_idx < row_size ) {
        shared_psp[tid] = 0.0;

        for (unsigned int i = 0; i < num_events[0]; i++) {
            int row_begin = spiked[i] * row_size;

            shared_psp[tid] += w[row_begin+col_idx];
        }

        g_%(target)s[col_idx] += shared_psp[tid];
        col_idx += gridDim.x * blockDim.x;
    }
}
""",
    'invoke_kernel': """
void proj%(id_proj)s_psp(RunConfig cfg, const long int t, const %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* num_events, %(conn_args_header)s %(kernel_args_header)s) {
    cu_proj%(id_proj)s_psp<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(
        t, dt, plasticity,
        /* pre-synaptic events */
        spiked, num_events,
        /* connectivity */
        %(conn_args_invoke)s
        /* kernel config */
        %(kernel_args_invoke)s
    );
}
""",
    'kernel_decl': """void proj%(id_proj)s_psp(RunConfig cfg, const long int t, const %(float_prec)s dt, bool plasticity, int *spiked, unsigned int* num_events, %(conn_args_header)s %(kernel_args_header)s);
""",
    'host_call': """
    if ( pop%(id_pre)s._active && (pop%(id_pre)s.spike_count > 0) && proj%(id_proj)s._transmission ) {
        int tpb = 32;
        int nb = int(ceil(double(proj%(id_proj)s.num_rows()) / double(tpb)));

        // compute psp
        proj%(id_proj)s_psp(
            RunConfig(nb, tpb, 0, proj%(id_proj)s.stream),
            t, dt, proj%(id_proj)s._plasticity,
            /* pre-synaptic events */
            pop%(id_pre)s.gpu_spiked, pop%(id_pre)s.gpu_spike_count,
            /* connectivity */
            %(conn_args)s
            /* kernel config */
            %(kernel_args)s
        );
    #ifdef _DEBUG
        cudaDeviceSynchronize();
        cudaError_t err_psp_proj%(id_proj)s = cudaGetLastError();
        if( err_psp_proj%(id_proj)s != cudaSuccess) {
            std::cout << "proj%(id_proj)s_psp (" << t << "): " << std::endl;
            std::cout << "   " << cudaGetErrorString(err_psp_proj%(id_proj)s) << std::endl;
        }
    #endif
    }
"""
}

conn_templates = {
    # connectivity representation
    'conn_header': "const %(idx_type)s post_size, const %(idx_type)s pre_size",
    'conn_call': "pop%(id_post)s.size, pop%(id_pre)s.size",
    'conn_kernel': "",

    # launch config
    'launch_config': launch_config,

    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,
    'delay': delay,

    #operations
    'rate_psp': None,
    'spike_transmission': {
        'event_driven': spike_event_transmission,
        'continuous': None
    },
    'synapse_update': {
        'global': None,
        'semiglobal': None,
        'local': None,
        'call': None
    }
}

conn_ids = {
    'local_index': "[j]",
    'semiglobal_index': '[rk_post]',
    'global_index': '[0]',
    'pre_index': '[rk_pre]',
    'post_index': '[rk_post]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_'
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np
from ANNarchy.core.Random import RandomDistribution
from ANNarchy.intern import Messages

######################
# Sparse matrices
######################

def sparse_random_matrix(pre, post, p, weight):
    """
    Returns a sparse (lil) matrix to connect the pre and post populations with the
    probability *p* and the value *weight*, either a constant or an ANNarchy random
    distribution object.
    """
    try:
        from scipy.sparse import lil_matrix
    except:
        Messages._warning("scipy is not installed, sparse matrices won't work")
        return None
    from random import sample
    W=lil_matrix((pre, post))
    for i in range(pre):
        k=np.random.binomial(post,p,1)[0]
        tmp = sample(range(post),k)
        W.rows[i]=list(np.sort(tmp))
        if isinstance(weight, (int, float)):
            W.data[i]=[weight]*k
        elif isinstance(weight, RandomDistribution):
            W.data[i]=weight.get_list_values(k)
        else:
            raise ValueError("sparse_random_matrix expects either a float or RandomDistribution object.")

    return W

def sparse_delays_from_weights(weight_matrix, delay):
    """
    Generates a delay matrix corresponding to the connectivity stored *weight_matrix*.
    """
    try:
        from scipy.sparse import lil_matrix
    except:
        Messages._warning("scipy is not installed, sparse matrices won't work")
        return None

    delay_matrix = lil_matrix(weight_matrix.get_shape())

    (rows,cols) = weight_matrix.nonzero()

    for r, c in zip(rows, cols):
        if isinstance(delay, (int, float)):
            delay_matrix[r,c] = delay
        elif isinstance(delay, RandomDistribution):
            delay_matrix[r,c] = delay.get_value()
        else:
            raise ValueError("sparse_random_matrix expects either a float or RandomDistribution object.")

    return delay_matrix

################################
## Performance Measurment
################################

def compute_delivered_spikes(proj, spike_events):
    """
    This function counts the number of delivered spikes for a given Projection and
    spike sequence.

    *Params*:

    * proj:             the Projection
    * spike_events:     the spike events per time step (the result of a spike recording)
    """
    nb_efferent_synapses = proj.nb_efferent_synapses()
    delivered_events = 0

    for neur_rank, time_steps in spike_events.items():
        if neur_rank in proj.pre.ranks and neur_rank in nb_efferent_synapses.keys():
            delivered_events += nb_efferent_synapses[neur_rank] * len(time_steps)

    return delivered_events

def compute_delivered_spikes_per_second(proj, spike_events, time_in_seconds, scale_factor=(10**6)):
    """
    This function implements a throughput metric for spiking neural networks.

    *Params*:

    * proj:             the Projection
    * spike_events:     the spike events per time step (the result of a spike recording)
    * time_in_seconds:  computation time used for the operation
    * scale_factor:     usually the throughput value gets quite large, therefore its useful to rescale the value. By default, we re-scale to millions of events per second.

    **Note**:

    In the present form the function should be used only if a single projection is recorded. If multiple Projections contribute to the
    measured time (*time_in_seconds*) you need to compute the value individually for each Projection using the *compute_delivered_spikes*
    method and the times for instance retrieved from a run using --profile.
    """
    num_events = compute_delivered_spikes(proj, spike_events)

    return (num_events / time_in_seconds) / scale_factor



# SpecificPopulation inheritances
from .InputArray import InputArray
from .PoissonPopulation import PoissonPopulation
from .SpikeSourceArray import SpikeSourceArray
from .SpikeTrains import HomogeneousCorrelatedSpikeTrains
from .TimedArray import TimedArray, TimedPoissonPopulation

# SpecificProjecion inheritances
from .CurrentInjection import CurrentInjection
from .DecodingProjection import DecodingProjection

__all__ = [
    'InputArray',
    'PoissonPopulation', 'SpikeSourceArray', 'TimedArray',
    'HomogeneousCorrelatedSpikeTrains', 'TimedPoissonPopulation',
    'DecodingProjection', 'CurrentInjection'
]


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern import Messages

class GlobalObjectManager:
    """
    Helper class to ensure, that the created *Constant* objects are globally accessible.
    """
    # singleton instance
    _instance = None
    # storage
    _objects = None

    def __init__(self):
        """
        Constructor.
        """
        pass

    def __new__(cls):
        """
        First call construction of the NetworkManager. No additional arguments are required.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._create_initial_state(cls)
        
        return cls._instance

    def _create_initial_state(self):
        """
        Initialize the container for globally available objects.

        Called either from __new__ or clear(). The first
        slot is reserved for the magic network.
        """
        self._objects = {
            'neurons': [],
            'synapses': [],
            'constants': [],
            'functions': []
        }

    def clear(self):
        """
        Remove all instantiated constants.
        """
        for obj in self._objects['constants']:
            del obj
        for obj in self._objects['neurons']:
            del obj
        for obj in self._objects['synapses']:
            del obj

        self._create_initial_state()

    ################################
    ## Neuron types
    ################################
    def add_neuron_type(self, neuron):
        self._objects['neurons'].append(neuron)

    def get_neuron_types(self):
        return self._objects['neurons']

    def num_neuron_types(self):
        return len(self._objects['neurons'])

    ################################
    ## Synapse types
    ################################
    def add_synapse_type(self, synapse):
        self._objects['synapses'].append(synapse)

    def get_synapse_types(self):
        return self._objects['synapses']

    def num_synapse_types(self):
        return len(self._objects['synapses'])

    ################################
    ## Constants
    ################################
    def add_constant(self, new_constant):
        """
        Add a constant to the list.
        """
        # avoid doublons
        for obj in self._objects['constants']:
            if obj.name == new_constant.name:
                Messages._error('the constant', new_constant.name, 'is already defined.')
        # add to global list of constants
        self._objects['constants'].append(new_constant)

    def list_constants(self):
        """
        Returns a list of all constants declared with ``Constant(name, value)``.
        """
        l = []
        for obj in self._objects['constants']:
            l.append(obj.name)
        return l

    def get_constant(self, name):
        """
        Returns the ``Constant`` object with the given name, ``None`` otherwise.
        """
        for obj in self._objects['constants']:
            if obj.name == name:
                return obj
        return None

    def get_constants(self):
        return self._objects['constants']

    def number_constants(self):
        return len(self._objects['constants'])

    ################################
    ## Functions
    ################################
    def add_function(self, function):
        name = function.split('(')[0]
        self._objects['functions'].append( (name, function) )

    def functions(name:str, network=None):
        net_id = 0 if network is None else network.id
        try:
            func = getattr(NetworkManager().cy_instance(net_id=net_id), 'func_' + name)
        except:
            Messages._error('call to', name, ': the function is not compiled yet.')

        return func

    def get_functions(self):
        return self._objects['functions']

    def number_functions(self):
        return len(self._objects['functions'])


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.ConfigManagement import get_global_config
from ANNarchy.intern import Messages

import numpy as np

class Dendrite :
    """
    A `Dendrite` is a sub-group of a `Projection`, gathering the synapses between the pre-synaptic population and a single post-synaptic neuron.

    It can not be created directly, only through a call to ``Projection.dendrite(rank)``:

    ```python
    dendrite = proj.dendrite(6)
    ```
    """
    def __init__(self, proj, post_rank, idx):

        self.post_rank = post_rank
        "Rank of the post-synaptic neuron."
        self.idx = idx
        self.proj = proj
        "Parent projection."
        self.pre = proj.pre

        self.target = self.proj.target

        self.attributes = self.proj.attributes
        self.parameters = self.proj.parameters
        self.variables = self.proj.variables

    @property
    def size(self) -> int:
        """
        Number of synapses.
        """
        if self.proj.cyInstance:
            return self.proj.cyInstance.dendrite_size(self.idx)
        return 0

    @property
    def pre_ranks(self) -> list[int]:
        """
        List of ranks of pre-synaptic neurons.
        """
        if self.proj.cyInstance:
            return self.proj.cyInstance.pre_rank(self.idx)
        return []

    def __len__(self):
        # Number of synapses.
        return self.size

    @property
    def synapses(self) -> "IndividualSynapse":
        """
        Iteratively returns the synapses corresponding to this dendrite.
        """
        for n in self.pre_ranks:
            yield IndividualSynapse(self, n)

    def synapse(self, pos:int|tuple[int]) -> "IndividualSynapse":
        """
        Returns the synapse coming from the corresponding presynaptic neuron.

        :param pos: can be either the rank or the coordinates of the presynaptic neuron
        :returns: `IndividualSynapse` wrapper instance.
        """
        if isinstance(pos, int):
            rank = pos
        else:
            rank = self.proj.pre.rank_from_coordinates(pos)

        if rank in self.pre_ranks:
            return IndividualSynapse(self, rank)
        else:
            Messages._error(" The neuron of rank "+ str(rank) + " has no synapse in this dendrite.")
            return None

    # Iterators
    def __getitem__(self, *args, **kwds):
        # Returns the synapse of the given position in the presynaptic population.
        # If only one argument is given, it is a rank. If it is a tuple, it is coordinates.
        
        if len(args) == 1:
            return self.synapse(args[0])
        return self.synapse(args)

    def __iter__(self):
        # Returns iteratively each synapse in the dendrite in ascending pre-synaptic rank order.
        for n in self.pre_ranks:
            yield IndividualSynapse(self, n)

    #########################
    ### Access to attributes
    #########################
    def __getattr__(self, name):
        # Method called when accessing an attribute.
        if name == 'proj':
            return object.__getattribute__(self, name)
        elif hasattr(self, 'proj'):
            if name == 'rank': # TODO: remove 'rank' in a future version
                Messages._warning("Dendrite.rank: the attribute is deprecated, use Dendrite.pre_ranks instead.")
                return self.proj.cyInstance.pre_rank(self.idx)
            
            elif name == 'pre_rank':
                return self.proj.cyInstance.pre_rank(self.idx)
            
            elif name == 'delay':
                if self.proj.uniform_delay == -1:
                    return [d*get_global_config('dt') for d in self.proj.cyInstance.get_dendrite_delay(self.idx)]
                else:
                    return self.proj.max_delay * get_global_config('dt')
            
            elif name == "w" and self.proj._has_single_weight():
                return self.proj.cyInstance.get_global_attribute(name, get_global_config('precision'))
            
            elif name in self.proj.attributes:
                # Determine C++ data type
                ctype = None
                for var in self.proj.synapse_type.description['variables']+self.proj.synapse_type.description['parameters']:
                    if var['name'] == name:
                        ctype = var['ctype']

                if name in self.proj.synapse_type.description['local']:
                    return self.proj.cyInstance.get_local_attribute_row(name, self.idx, ctype)
                elif name in self.proj.synapse_type.description['semiglobal']:
                    return self.proj.cyInstance.get_semiglobal_attribute(name, self.idx, ctype)
                else:
                    return self.proj.cyInstance.get_global_attribute(name, ctype)
            else:
                return object.__getattribute__(self, name)
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        # Method called when setting an attribute.
        if name == 'proj':
            object.__setattr__(self, 'proj', value)
        elif name == 'attributes':
            object.__setattr__(self, 'attributes', value)
        elif hasattr(self, 'proj'):
            if name in self.proj.attributes:
                # Determine C++ data type
                ctype = None
                for var in self.proj.synapse_type.description['variables']+self.proj.synapse_type.description['parameters']:
                    if var['name'] == name:
                        ctype = var['ctype']

                if name in self.proj.synapse_type.description['local']:
                    if isinstance(value, (np.ndarray, list)):
                        self.proj.cyInstance.set_local_attribute_row(name, self.idx, value, ctype)
                    else:
                        self.proj.cyInstance.set_local_attribute_row(name, self.idx, value * np.ones(self.size), ctype)

                elif name in self.proj.synapse_type.description['semiglobal']:
                    self.proj.cyInstance.set_semiglobal_attribute(name, self.idx, value, ctype)

                else:
                    # HD: will break the execution of the program
                    Messages._error("Projection attributes marked as *projection* should not be updated through dendrites.")
            else:
                object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def set(self, value:dict) -> None:
        """
        Sets the value of a parameter/variable of all synapses.

        Example:

        ```python
        dendrite.set( { 'tau' : 20, 'w'= Uniform(0.0, 1.0) } )
        ```

        :param value: a dictionary containing the parameter/variable names as keys.
        """
        for key, value in value.items():
            # sanity check and then forward to __setattr__
            if key in self.attributes:
                setattr(self, key, value)
            else:
                Messages._error("Dendrite has no parameter/variable called", key)

    def get(self, name:str) -> float:
        """
        Returns the value of a variable/parameter.

        Example:

        ```python
        dendrite.get('w')
        ```

        :param name: name of the parameter/variable.
        :returns: a single value.
        """
        if name == 'rank':
            Messages._warning("Dendrite.get('rank'): the attribute is deprecated, use Dendrite.pre_ranks instead.")
            return self.proj.cyInstance.pre_rank(self.idx)
        elif name == 'pre_ranks':
            return self.proj.cyInstance.pre_rank(self.idx)
        elif name in self.attributes:
            return getattr(self, name)
        else:
            Messages._error("Dendrite has no parameter/variable called", name)


    #########################
    ### Formatting
    #########################
    def receptive_field(self, variable:str='w', fill:float=0.0) -> np.array:
        """
        Returns the given variable as a receptive field.

        A Numpy array of the same geometry as the pre-synaptic population is returned. 
        Non-existing synapses are replaced by zeros (or the value ``fill``).

        :param variable: name of the variable (default = 'w')
        :param fill: value to use when a synapse does not exist (default: 0.0).
        :returns: an array.
        """
        values = getattr(self.proj.cyInstance, 'get_dendrite_'+variable)(self.idx)
        pre_ranks = self.proj.cyInstance.pre_rank( self.idx )

        m = fill * np.ones( self.pre.size )
        m[pre_ranks] = values

        return m.reshape(self.pre.geometry)


    #########################
    ### Structural plasticity
    #########################
    def create_synapse(self, rank:int, w:float=0.0, delay:float=0) -> None:
        """
        Creates a synapse for this dendrite with the given pre-synaptic neuron.

        :param rank: rank of the pre-synaptic neuron
        :param w: synaptic weight.
        :param delay: synaptic delay.
        """
        if not get_global_config('structural_plasticity'):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not add the synapse.')
            return

        if self.proj.cyInstance.dendrite_index(self.post_rank, rank) != -1:
            Messages._error('the synapse of rank ' + str(rank) + ' already exists.')
            return

        # Set default values for the additional variables
        extra_attributes = {}
        for var in self.proj.synapse_type.description['parameters'] + self.proj.synapse_type.description['variables']:
            if not var['name'] in ['w', 'delay'] and  var['name'] in self.proj.synapse_type.description['local']:
                if not isinstance(self.proj.init[var['name']], (int, float, bool)):
                    init = var['init']
                else:
                    init = self.proj.init[var['name']]
                extra_attributes[var['name']] = init

        try:
            self.proj.cyInstance.add_synapse(self.post_rank, rank, w, int(delay/get_global_config('dt')), **extra_attributes)
        except Exception as e:
            Messages._print(e)

    def create_synapses(self, ranks:list[int], weights:list[float]=None, delays:list[float]=None) -> None:
        """
        Creates a synapse for this dendrite with the given pre-synaptic neurons.

        :param ranks: list of ranks of the pre-synaptic neurons.
        :param weights: list of synaptic weights (default: 0.0).
        :param delays: list of synaptic delays (default = dt).
        """
        if not get_global_config('structural_plasticity'):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not add the synapse.')
            return

        # No user-side init
        if weights is None:
            weights = [0.0] * len(ranks)

        if delays is None:
            delays = [0] * len(ranks)

        # Collect other attributes than w/delay
        extra_attribute_names = []
        for var in self.proj.synapse_type.description['parameters'] + self.proj.synapse_type.description['variables']:
            if not var['name'] in ['w', 'delay'] and  var['name'] in self.proj.synapse_type.description['local']:
                extra_attribute_names.append[var['name']]

        # Create the synapses
        for rank, w, delay in zip(ranks, weights, delays):
            if self.proj.cyInstance.dendrite_index(self.post_rank, rank) != -1:
                Messages._error('the synapse of rank ' + str(ranks) + ' already exists.')
                return

            # Set default values for the additional variables
            extra_attributes = {}
            for var in extra_attribute_names:
                if not isinstance(self.proj.init[var], (int, float, bool)):
                    init = var['init']
                else:
                    init = self.proj.init[var]
                extra_attributes[var] = init

            try:
                self.proj.cyInstance.add_synapse(self.post_rank, rank, w, int(delay/get_global_config('dt')), **extra_attributes)
            except Exception as e:
                Messages._print(e)

    def prune_synapse(self, rank:int) -> None:
        """
        Removes the synapse with the given pre-synaptic neuron from the dendrite.

        :param rank: rank of the pre-synaptic neuron
        """
        if not get_global_config('structural_plasticity'):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not remove the synapse.')
            return

        if not rank in self.pre_ranks:
            Messages._error('the synapse with the pre-synaptic neuron of rank ' + str(rank) + ' did not already exist.')
            return

        self.proj.cyInstance.remove_synapse(self.post_rank, rank)

    def prune_synapses(self, ranks:list[int]):
        """
        Removes the synapses which belong to the provided pre-synaptic neurons from the dendrite.

        :param ranks: list of ranks of the pre-synaptic neurons.
        """
        if not get_global_config('structural_plasticity'):
            Messages._error('"structural_plasticity" has not been set to True in setup(), can not remove the synapse.')
            return

        for rank in ranks:
            self.prune_synapse(rank)

class IndividualSynapse :

    def __init__(self, dendrite, rank):
        self.dendrite = dendrite
        self.rank = rank

    def __getattr__(self, name):
        " Method called when accessing an attribute."
        if name in ['dendrite', 'attributes', 'rank']:
            return object.__getattribute__(self, name)

        elif name in self.dendrite.proj.synapse_type.description['attributes']:
            # Determine C++ data type
            ctype = None
            for var in self.dendrite.proj.synapse_type.description['variables']+self.dendrite.proj.synapse_type.description['parameters']:
                if var['name'] == name:
                    ctype = var['ctype']

            if name in self.dendrite.proj.synapse_type.description['local']:
                return self.dendrite.proj.cyInstance.get_local_attribute(name, self.dendrite.idx, self.rank, ctype)
            elif name in self.dendrite.proj.synapse_type.description['semiglobal']:
                return self.dendrite.proj.cyInstance.get_semiglobal_attribute(name, self.dendrite.idx, ctype)
            else:
                return self.dendrite.proj.cyInstance.get_global_attribute(name, ctype)

        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        " Method called when setting an attribute."
        if name in ['dendrite', 'attributes', 'rank']:
            object.__setattr__(self, name, value)

        elif name in self.dendrite.proj.synapse_type.description['attributes']:
            # Determine C++ data type
            ctype = None
            for var in self.dendrite.proj.synapse_type.description['variables']+self.dendrite.proj.synapse_type.description['parameters']:
                if var['name'] == name:
                    ctype = var['ctype']

            if name in self.dendrite.proj.synapse_type.description['local']:
                self.dendrite.proj.cyInstance.set_local_attribute(name, self.dendrite.idx, self.rank, value, ctype)
            elif name in self.dendrite.proj.synapse_type.description['semiglobal']:
                self.dendrite.proj.cyInstance.set_semiglobal_attribute(name, self.dendrite.idx, value, ctype)
            else:
                self.dendrite.proj.cyInstance.set_global_attribute(name, value, ctype)

        else:
            object.__setattr__(self, name, value)
